// sketch for fusor control arduino uno, slaved to raspi
// Copyright 2016 Doug Coulter, license GPLv2 - share and enjoy, but do share back
// Arduino IDE 1.6.10 on ARM used for building and testing
/*
 * Command format is one per line and either just an address (ascii) followed 
 * by a command letter and a newline (say 5a\n in this case would make us send ack\n)
 * or an address, letter, and numeric value (in ascii also) followed by \n
 * We insist on an address as we might be on a shared serial bus
 * We also insist on a command being contiguous and complete inside MAX_TIME which
 * is currently set generously for the serial speed.
 *  Note this version (1.6.10) allows setting serial speed up to 500k baud, but only
 *  actually supports 250k in the serial monitor.  We're using less for PC or Pi
 *  "works everywhere" compatibility.  NB - on first run you have to set the eeprom,
 *  I don't know how from the compiler.
 *  
 *  The usual caveat - anything I marked with @@@ I thought was either dodgy or incomplete.
 */

// pi sees us as:/dev/serial/by-id/usb-1a86_USB2.0-Serial-if00-port0

#include <MsTimer2.h> // way to get ms rez interrupt for gas in solenoid (takes an unsigned long...we won't need that)
#include <EEPROM.h>  // save some tuning variables between runs

//############################# Pinouts #########################################
// pin D3 is clap input - go on falling edge
// pin D5 is the fast counter input - falling edge counts, 32 bit total count kept
// we report at slow loop speeds difference from last time, dunno if wraps right @@@
#define DA_PIN 6
#define HVOFF_PIN 8
#define IONOK_PIN 9
#define GAS_IN_PIN 10
#define GAS_OUT_PIN 11
#define LED_PIN 13
// a0 Main Current Set
// a1 Main Current
// a2 Main Pressure (PKR-251)
// a3 Ion mA 
// other a/d not used yet - but we have the speed and space for them all


//############################# other defines ####################################
// Want to get 10 fastloops in per slowloop, but save some time so we never have 
// milliscond creep (goes with a "4" below - the actual routine takes around 3ms)
#define FAST_LOOP_TIME 9
#define SLOW_LOOP_TIME 100
#define SLOW_LOOP_MARGIN 4

// current spot we're wearing out in eeprom
#define EEPROM_ADDRESS 100

//############################### RAM  ###########################################

// struct to save/restore with eeprom
struct Persist {
  byte DaDefault;
  byte GasOutMS;
  byte GasInMS;
};
Persist Tune; // same thing as above, in ram
// things touched by ISRs marked volatile (self-defense hungarian notation)
volatile boolean Hostenable; // various flags that can turn us off, this from host
volatile boolean HVOK; // this one from main HV supply
volatile boolean IonOK; // this one from ion supply
boolean LEDState;
byte DAOut; // value for D/A to put out (ion power control)

int i;  // iterator for background stuff
// loops variables
unsigned long fastmilliz; // z^-1 values from last time
unsigned long slowmilliz;
// data
#define ADCHNLS 4  // number of in-use a/d channels (starts at ch0)
int AD[ADCHNLS];        // values read from the a/d
#define fADi 0.1      // pretty fast filter at .2
#define fADa (1.0-fADi) // make sure filter coeficients add to exactly one in float @@@ parens required here!
float fAD[ADCHNLS];   // for iir average
// we'll think of something to do with a fast hardware counter I'm sure
volatile unsigned long c1counts;    // total counts from (hardware) counter 1, d5 pin
// top half incremented by timer overflow
volatile unsigned long c1z; // previous count for subtraction
volatile unsigned long curcount; // what we'll report, this - last total
unsigned long claptime;    // millis time at clapboard signal (for time sync with other units)
volatile unsigned long false_claps;

// stuff for command state machine below
// we use an address even though this one is USB - someday might be plain serial and shared
#define MY_ADDRESS '5'  // could be any byte, but we're using all-ascii for now, and '\n' for end of command
#define CMD_BUF_SIZE 32 // in theory, two of these would fit in the uno USB buffer at once.  Probably don't need this much IRL
#define MAX_TIME 20 // milliseconds to timeout and ignore if incomplete command (3ms would do for 115200 and fast commands)
byte Command[CMD_BUF_SIZE]; // command buffer for incoming commands
byte CommandIndex; // where am I in above buffer
byte CommandState; // state variable
unsigned long CommandTime; // time at start of command recieve
////// done with command-slave state machine stuff
byte HostFlags;

//################################################################
//#####################   Code   #################################
//################################################################
////////////////////////////////////////////////
void initRam() // used at startup and other times - be clean!
{
  // note things like bool, int, or float combined on same line for speed/efficiency
 Hostenable = IonOK = HVOK = false; // come up "off"
 CommandIndex = CommandState = HostFlags = 0;
 fastmilliz = slowmilliz = 0;
 DAOut = 0;
 for (i=0; i<ADCHNLS;i++) {fAD[i] = 0.0;} // long term averages need zeroed
 c1counts = c1z = claptime = false_claps = 0;
 EEPROM.get(EEPROM_ADDRESS,Tune); // tuning saved from previous runs
}
////////////////////////////////////////////////
void t1set() // setup timer1 as a counter that interrupts on overflow
{
  TCCR1A = 0; // turn off all the weird compare/capture modes etc
  TCCR1B = 6; // clock on falling edge (7 for rising)
  TIFR1 = _BV(TOV1); // clear initial overflow - shouldn't need to, but do.
  TCNT1 = 0; //zero out timer
  TIMSK1 = _BV(TOIE1);// sets the timer overflow interrupt enable bit
  }
////////////////////////////////////////////////
void CloseValves() // close all the gas valves and turn off timer
{
 digitalWrite(GAS_IN_PIN,LOW); // do first as this is the fast one
 digitalWrite(GAS_OUT_PIN,LOW); // might as well do both all the time (suspenders)
 MsTimer2::stop(); // once should get it done
}
//////////////////////////////////////////////// SETUP ##################
void setup() { // runs at reset
 pinMode(LED_PIN, OUTPUT); // set pin modes
 pinMode(DA_PIN,OUTPUT); // pwm output for ion supply
 pinMode(GAS_IN_PIN,OUTPUT); // gas inlet solenoid drive
 pinMode(GAS_OUT_PIN,OUTPUT); // outlet solenoid drive
// pins default to input, and to zero, so a few lines are not needed here

 // set variables
 t1set(); // don't forget this if you get strange count outputs!
 initRam();  // basically, make sure valid and "all turned off"
 attachInterrupt(1,setClap,FALLING); // wait for signal from host on pin d3 to begin to do time-sync
 Serial.begin(115200); // use USB serial for host comm
 Serial.println ("gasion setup done FC version");

}
////////////////////////////////////////////////
//////// manual isr setup, you bet, saves a whole library
// actual isr for hardware counter 1
ISR(TIMER1_OVF_vect)
{
 c1counts += 0x00010000; // add one to top half of longword
}
//////////////////////////////////////////////// 
///////////### UTILITIES ###//////////////////// 
//////////////////////////////////////////////// 
void readCounts()
{ // just fast hardware counter (d5 pin), interrupts us only every 65k counts
/* Disable interrupts */
noInterrupts();
/* Read TCNT1 into c1counts low part */
c1counts &= 0xffff0000;
c1counts |= TCNT1; // from the hardware register @@@ there's a better capture register
curcount = c1counts - c1z; // just the new ones for this time interval
c1z = c1counts; // for next time
/* Restore global interrupt flag */
interrupts(); 
}
////////////////////////////////////////////////
void setion ()
{ // input level is a char for pwm duty 0-204 is 0-4v or 0.019607843v per unit
  // 153 ~ 3v, 102 ~ 2v, 51 ~ 1v
 analogWrite(DA_PIN,DAOut); // set volts for ion supply (0-4v is 0-40kv but we won't use full range)
}
////////////////////////////////////////////////
void readAD() // 10 bit a/d.  Sum of 4 samples is 12 bits long (and *almost* 12 bits good with noise to dither)
{
 byte chnl;  // Analog input pin 
 byte pass;

  for (chnl=0; chnl < ADCHNLS; chnl++) {AD[chnl] = 0; } // clear out old
  for (pass=0; pass < 4; pass++)
  {
   for (chnl=0; chnl < ADCHNLS; chnl++)
   {
    AD[chnl]  += analogRead(chnl); // delay required?  No.
   }
  } // for pass
}
////////////////////////////////////////////////
/*float Str2Flt()
{ // parse string in Command[2] through end into float, return the float
 String numstring = ""; // to hold input number
 float value; // result

 i=2; // shared var, iterator/index here, 0 and 1 are unit number and command
 while (Command[i] != '\n') // \n is command end character
 {
//  Serial.print(char(Command[i])); // debug
  numstring += char(Command[i++]); // really does need cast to char
 }
// Serial.println(":done");
 value = numstring.toFloat();
 return value; // send it to caller
}*/ // tested and working, not currently in use
////////////////////////////////////////////////
byte Str2Byte()
{ // parse string in Command[2] through end into byte, return the byte
 String numstring = ""; // to hold input number
 int value; // result

 i=2; // shared var, iterator/index here, 0 and 1 are unit number and command
 while (Command[i] != '\n') // \n is command end character
 {
//  Serial.print(char(Command[i])); // debug
  numstring += char(Command[i++]); // really does need cast to char
 }
// Serial.println(":done");
 value = numstring.toInt();
 if (value < 0) {value = 0;} // limit to reasonable values for a byte
 if (value > 255) {value = 255;}
 return (byte)value; // send it to caller
}
////////////////////////////////////////////////

//////////////////////////////////////////////// 
//////////### COMMANDS FROM HOST ###////////////
////////////////////////////////////////////////
void soft_reset () // software reset, actually a hard RST
{ // so we can reset timers and so on under host control
 asm volatile ("  jmp 0"); 
} // this really is a hard reset, but coming from software
////////////////////////////////////////////////
void setClap ()
{
  if (0 == claptime)
  {
   claptime = millis();  // @@@ make sure counters are clear too?
  } else
  {
    false_claps++; // count errors
  }
}
////////////////////////////////////////////////
void CommandMachine()
{
 switch (CommandState)
 {
  case 0: // looking for work to do
  if (Serial.available())
  {
   CommandTime = millis(); // for later possible timeout
   CommandState = 1; // get bytes till done or timeout
  }
  break;

  case 1: // getting command
  while (Serial.available())
  {
   Command[CommandIndex] = Serial.read(); // get a byte if available
//   Serial.print ("received:"); // debug
//   Serial.println(Command[CommandIndex]); // debug
   
   if (Command[CommandIndex] == '\n') // single quotes, else compiler thinks pointer to string
    { // got a full command, go parse it
     CommandState = 2;
     return; // do it on next roundy round of loop() 
    }
   CommandIndex++; // bump the position
   if ((CommandIndex >= CMD_BUF_SIZE) || ((CommandTime + MAX_TIME) < millis())) // enough parens?
   { // error of some kind, reset state machine - SOMETHING SCREWED UP
    CommandIndex = 0;
    CommandState = 0; // ignore what might have been garbage
//    Serial.println ("Command error"); // debug
   }
  } // end while
  break;
  
  case 2: // parse command
   if (Command[0] != MY_ADDRESS) // is this even for us?
   { // no, ditch it
    CommandIndex = 0;
    CommandState = 0;
    return; // forget it ever happened "LA LA LA I CAN'T HEAR YOU IF YOU CAN'T SAY MY NAME"
   }
   // It might have been for us, any legality tests might go after this comment.
   // By convention, address of unit is first byte, command is next, data might follow, should be enough.
   // For time consuming or big stuff, use a subroutine call instead of inlining it like this example.
   // Commands are all-ascii for convienience of host (and we slaves).
   switch (Command[1])
    {
     case '!': // software reset
        soft_reset(); // clobber the world and start over
     break;
     case 'a': // host requests an ack
        Serial.println("ack");  // should work
     break;
     case 'c': // fake a clapboard
      if (0 == claptime)
      {
       claptime = millis();
      } else 
      {
       false_claps++; 
      }
     break;
     case 'd': // report our "junk"
      Debug(); // emit debug output
     break;
     case 'e': // enable gas/ion control, etc
      Hostenable = true;
      DAOut = Tune.DaDefault; // @@@ debug override for the moment
      HostFlags |= 4; // report ion voltage status
     break;
     case 'E': // turn ion off
     Hostenable = false;
     DAOut = 0;
     HostFlags &= 0xfb; // clear ion voltage bit
     break;
     case 'i': // set gas in duration
     Tune.GasInMS = Str2Byte();
     break;
     case 'I': // actually let gas in
      digitalWrite(GAS_IN_PIN,HIGH); // do the deed
      HostFlags |= 2; // note it for reporting
      MsTimer2::set(Tune.GasInMS, CloseValves); // whatever is default gas-in time
      MsTimer2::start(); // close valves at timeout
     break;
     case 'o': // input gas out timing, parse and store
     Tune.GasOutMS = Str2Byte();
     break;
     case 'O':
      digitalWrite(GAS_OUT_PIN,HIGH);
      HostFlags |= 1; // note it
      MsTimer2::set(Tune.GasOutMS, CloseValves); // whatever is default gas-out time
      MsTimer2::start(); // close valves at timeout    
     break;
     case 'k': //host kills us @@@ add software reset?
     Hostenable = false;
     EEPROM.put(EEPROM_ADDRESS,Tune); // stash the tuning variable struct
     initRam();
     break;     
     case 's': // save our tuning in eeprom
      EEPROM.put(EEPROM_ADDRESS,Tune); // stash the tuning variable struct
     break;
     case 'v': // da default. 200 ~ 4v "Voltage, since d is used already"
     Tune.DaDefault = Str2Byte();
     DAOut = Tune.DaDefault; // actually do it
     break;
    }    
  // we are done with this one -- note no default case needed.  If it's crap, we ignored it.
     CommandIndex = 0;
     CommandState = 0;
//     Serial.println("end state 2"); // debug 
   break; // end state 2
 } // end switch state machine
} // end command state machine routine

////////////////////////////////////////////////
void DataOut()
{ // data inputs we collected for logging back to the host

// Serial.print (Serial.availableForWrite());
// Serial.print ("\t");
 if (0 != false_claps)
 {
  Serial.print("# iongas false claps: ");
  Serial.println(false_claps);
  false_claps = 0;
 }
 Serial.print (millis() - claptime);
 byte i;
 for (i=0; i < ADCHNLS; i++)
 {
  Serial.print ("\t");  // some delimiter
  Serial.print ((int)fAD[i]); // get the good lowpassed version
 } 
 Serial.print ("\t");
 Serial.print(curcount); // new (only) counts from hardware counter


// now my own outputs
 Serial.print ("\t");
 Serial.print(DAOut); // current d/a output duty cycle (of 256)
 Serial.print ("\t");
 Serial.print (HostFlags);
  // was gas let in (2)?  Gas going out (1)?  Things ok? (upper nybble)
 Serial.println(); // EOL 
 HostFlags &= 0xfc; // clear lower bits
}
////////////////////////////////////////////////
void Debug ()
{
 
 Serial.print ("#Debug\t millis:");
 Serial.print (millis());
 Serial.print ("\tclaptime:");
 Serial.print (claptime);
 Serial.print ("\tPersist.DaDefault:");
 Serial.print (Tune.DaDefault);
 Serial.print ("\tGasOutMS:");
 Serial.print (Tune.GasOutMS);
 Serial.print ("\tGasInMS:");
 Serial.print (Tune.GasInMS);

 Serial.println();

}
//################################################################
void fastloop()
{ // do things at the fast rate, not interrupt driven for now
 fastmilliz = millis();
 readAD(); // get a/d data at the fast rate
 for (i=0;i<ADCHNLS;i++)
 {
  fAD[i] = (fAD[i] * fADa) + (fADi * float(AD[i])); // iir filter each channel in float
 }
  IonOK = digitalRead(IONOK_PIN); // poll some stuff
  HVOK = !digitalRead(HVOFF_PIN);
  setion(); // pwm output to control ion source
  // approx 3 ms per loop at this point
}
//################################################################
void slowloop()
{ // the slow loop - report status to host and so on
 slowmilliz = millis();
 readCounts(); // results go to c1counts unsigned long 
 if (0 != claptime)
 { // don't do reporting etc until told to begin
  DataOut();  // send stuff to host
 } // if clapped at
} // slow loop
//################################################################
//                    MAIN
//################################################################

void loop() 
 { // main code here, to run repeatedly:

  CommandMachine();  // do things at request of host - as fast as we can get to it
  if (millis() - slowmilliz >= SLOW_LOOP_TIME)   { slowloop();}// run things at a more sane rate 
  if ((millis() - fastmilliz >= FAST_LOOP_TIME) && 
     (millis() - slowmilliz < (SLOW_LOOP_TIME-SLOW_LOOP_MARGIN))) { fastloop();} 
     // run things that need speed
///@@@ make sure if you change the above, or how long fastloop() takes, that you adjust the other things (4, the the slow and fast loop times) so we don't
// have milliscond creep. This is an especial pain for the counter data.
  digitalWrite(LED_PIN,LEDState); // show time through loop
  LEDState  = !LEDState; // toggle it  
 }
